#!/usr/bin/env python3
# -*- encoding: utf-8 -*-

import logging
from copy import copy
from functools import partial
from itertools import chain, combinations, product

import six

from .patterns import dollar_prefix, pop_dollar_prefix, _convert_names_with_underlines_to_dots

logger = logging.getLogger(__name__)


class MongoAggregation(list):

    def __init__(self, pipeline='', collection='', allowDiskUse=False):
        self.collection = collection
        self.allowDiskUse = allowDiskUse
        self.actual_fields = set()
        self.pipeline = pipeline if pipeline else []

    def aggregate(self, collection='', allowDiskUse=False, as_list=False, collation=None):
        if collection:
            self.collection = collection
        if self.collection.__class__.__name__ == 'TopLevelDocumentMetaclass':
            self.collection = self.collection.objects
        if allowDiskUse:
            self.allowDiskUse = allowDiskUse
        if self.collection.__class__.__name__ != 'QuerySet' and not self.collection:
            logger.error('Агрегация невозможна: не указана коллекция')
            return
        aggregate = partial(self.collection.aggregate, allowDiskUse=self.allowDiskUse, collation=collation)
        if self.collection.__class__.__name__ == 'QuerySet':
            result = aggregate(*self.pipeline)
        else:
            result = aggregate(self.pipeline)
        return list(result) if as_list else result

    def append(self, object=None, *args):
        if not object: object = []
        if isinstance(object, list):
            self.pipeline.extend(object)
        else:
            self.pipeline.append(object)
        if args:
            for arg in args:
                self.pipeline.append(arg)
        return self

    def extend(self, object=None, *args):
        self.append(object, *args)

    def count(self, **kwargs):
        self.pipeline.append({'$count': 'count'})
        count = next(self.aggregate(**kwargs), {}).get('count', 0)
        # Revert last stage
        self.revert_last_stage()
        return count

    def match(self, *args, **kwargs):
        if not args and not kwargs:
            return self
        if args:
            for arg in args:
                self.pipeline.append({
                    '$match': arg
                })
        kwargs = _convert_names_with_underlines_to_dots(kwargs, convert_operators=True)
        if args and kwargs:
            self.pipeline[-1]['$match'].update(kwargs)
        elif kwargs:
            self.pipeline.append({
                '$match': kwargs
            })
        self._add_to_actual_fields([
            x for x in self.last_stage.statement.keys()
            if not x.startswith('$')
        ])
        return self

    def lookup_unwind(self, collection, local_field='_id', as_field='', foreign_field='_id',
                      preserveNullAndEmptyArrays=True):
        if not as_field:
            as_field = local_field
        self.lookup(collection, local_field, as_field, foreign_field),
        self.unwind(as_field, preserveNullAndEmptyArrays)
        return self

    def lookup(self, collection, local_field='_id', as_field='', foreign_field='_id'):
        if not as_field:
            as_field = local_field
        self.pipeline.append({"$lookup": {
            'from': collection,
            'foreignField': foreign_field,
            'localField': local_field,
            'as': as_field,
        }})
        self._add_to_actual_fields(local_field, ignore_if_theres_children=True)
        self._add_to_actual_fields(as_field, ignore_if_theres_children=True)
        return self

    def unwind(self, field, preserveNullAndEmptyArrays=False):
        field = dollar_prefix(field)
        if preserveNullAndEmptyArrays:
            self.pipeline.append({'$unwind': {'path': field, 'preserveNullAndEmptyArrays': True}})
        else:
            self.pipeline.append({'$unwind': field})
        self._add_to_actual_fields(field, ignore_if_theres_children=True)
        return self

    def order_by(self, *args, **kwargs):
        return self.sort(*args, **kwargs)

    def sort(self, *args, **kwargs):
        def _prepare_str_order_rule(order_rule):
            if not isinstance(order_rule, six.string_types):
                return order_rule
            sign = -1 if order_rule.startswith('-') else 1
            sort_field = order_rule.strip('-+')
            return {sort_field: sign}

        kwargs = _convert_names_with_underlines_to_dots(kwargs)
        # Складываем все в args
        args = list(args)
        if kwargs:
            args.append(kwargs)

        if not args: return self
        # Делаем project
        for order_rule in args:
            order_rule = _prepare_str_order_rule(order_rule)
            self.pipeline.append({"$sort": order_rule})
            self._add_to_actual_fields(order_rule.keys())
        return self

    def skip(self, offset=0):
        if not offset: return self
        self.pipeline.append({"$skip": offset})
        return self

    def limit(self, limit=0):
        if not limit: return self
        self.pipeline.append({"$limit": limit})
        return self

    def project(self, *args, **kwargs):
        kwargs = _convert_names_with_underlines_to_dots(kwargs)
        # Складываем все в args
        if args and kwargs:
            args[-1].update(kwargs)
        elif kwargs:
            args = (kwargs,)
        if not args: return self

        # Делаем project
        for arg in args:
            self.pipeline.append({'$project': arg})
        # Заполняем актуальные поля
        self._set_actual_fields(arg.keys())
        return self

    def _get_fields_without_children(self, fields):
        parents = copy(fields)
        for one, another in combinations(fields, 2):
            is_children = another.startswith(f'{one}.')
            if is_children:
                parents.discard(another)
                continue
            is_children = one.startswith(f'{another}.')
            if is_children:
                parents.discard(one)
                continue
        return parents

    @staticmethod
    def _str_to_list(variable):
        return variable.split(',') if isinstance(variable, str) else variable

    def _set_actual_fields(self, *fields):
        """Устанавливает список актуальных полей"""
        self.actual_fields = set(chain(*fields))

    def _add_to_actual_fields(self, fields, ignore_if_theres_children=True):
        """Добавляет поля в список актуальных полей"""
        _fields = set(self._str_to_list(fields))
        if ignore_if_theres_children:
            _fields = [
                field
                for field in _fields
                if pop_dollar_prefix(field) not in self.get_parents(self.actual_fields, all_levels=True)
            ]
        for field in _fields:
            self.actual_fields.add(pop_dollar_prefix(field))

    @staticmethod
    def get_parents(fields, all_levels=False, upper_level=True):
        if isinstance(fields, str):
            fields = set(fields.split(','))

        # Самый нижний уровень
        if not all_levels and not upper_level:
            return {field.split('.')[0] for field in fields}

        # Все уровни и верхний
        parents = set()
        for field in fields:
            splitted = field.split('.')
            for i in range(len(splitted)):
                splitted.pop()
                if not splitted: break
                parents.add('.'.join(splitted))
                # Только верхний уровень
                if not all_levels: break
        return parents

    def replace_root(self, expression):
        """Replaces the input document with the specified document.
        The operation replaces all existing fields in the input document, including the _id field.
        From MongoDB version 3.4."""
        if isinstance(expression, six.string_types):
            expression = dollar_prefix(expression)
        self.pipeline.append({'$replaceRoot': {'newRoot': expression}})
        return self

    def add_fields(self, **kwargs):
        """Adds new fields to documents.
        $addFields outputs documents that contain all existing fields from the input documents and newly added fields.
        From MongoDB version 3.4
        Starting in version 4.2, MongoDB adds a new aggregation pipeline stage $set that is an alias for $addFields.
        """
        kwargs = _convert_names_with_underlines_to_dots(kwargs)
        self.pipeline.append({'$addFields': kwargs})
        self._add_to_actual_fields(kwargs.keys())
        return self

    def set(self, **kwargs):
        """Adds new fields to documents.
        $set outputs documents that contain all existing fields from the input documents and newly added fields.
        From MongoDB version 4.2."""
        kwargs = _convert_names_with_underlines_to_dots(kwargs)
        self.pipeline.append({'$set': kwargs})
        self._add_to_actual_fields(kwargs.keys())
        return self

    def smart_project(self, include_fields='', exclude_fields='', include_all_by_default=True, *args, **kwargs):
        """Custom realization of $addFields for an older versions of MongoDB. Bases on $project stage."""
        if not include_all_by_default:
            self.actual_fields = set()

        if include_fields:
            include_fields = set(include_fields.split(','))
            parent_fields = self.get_parents(include_fields, True)
            self.actual_fields -= parent_fields
            self.actual_fields.update(include_fields)

        # Проверяем обращение к полям объекта с помощью __ в kwargs
        kwargs = _convert_names_with_underlines_to_dots(kwargs)

        # Складываем kwargs в args
        args = self._add_kwargs_to_args(args, kwargs)

        if exclude_fields:
            exclude_fields = set(exclude_fields.split(','))

        if any(args) or include_fields:
            # Exclude fields may contain some information of what fields to include (calculate by contradiction)
            if exclude_fields:
                # Add parents first
                # actual_fields = {'count'}
                # exclude_field = 'menu.elements.option'
                # Parent: 'menu.elements'
                # Result: actual_fields = {'count', 'menu.elements'}
                for excl_field in exclude_fields:
                    self.actual_fields.update(self.get_parents(excl_field, True))
                for excl_field in exclude_fields:
                    # Remove the field itself
                    self.actual_fields.discard(excl_field)
                    # Remove all children
                    # actual_fields = {'menu.elements.option', 'menu.elements.count', 'count'}
                    # exclude_field = 'menu.elements'
                    # Children: 'menu.elements.option', 'menu.elements.count'
                    # Result: actual_fields = {'count'}
                    child_fields = {
                        act_field for act_field in self.actual_fields
                        if '{}.'.format(excl_field) in act_field
                    }
                    self.actual_fields -= child_fields

            paths = self._get_stage_hierarchy_fields(args[-1])
            if not paths and self.actual_fields:
                prolong_fields = self.actual_fields
            else:
                prolong_fields = set([actual_field
                                      for new_field, actual_field in product(paths, self.actual_fields)
                                      if actual_field != new_field
                                      and not actual_field.startswith(f'{new_field}.')
                                      and not new_field.startswith(f'{actual_field}.')])
            prolong_fields = self._get_fields_without_children(prolong_fields)
            for field in prolong_fields:
                args[-1].setdefault(field, 1)

            # Делаем project
            if args:
                for arg in args:
                    self.pipeline.append({'$project': arg})
                # Заполняем актуальные поля
                self._set_actual_fields(self._get_stage_hierarchy_fields(arg))

        # Then esclude fields and children
        if exclude_fields:
            # !!! This is the same section, we double it because it may save some fields in earlier projections
            # Add parents first
            for excl_field in exclude_fields:
                self.actual_fields.update(self.get_parents(excl_field, True))
            for excl_field in exclude_fields:
                # Remove the field itself
                self.actual_fields.discard(excl_field)
                # Remove all children
                child_fields = {
                    act_field for act_field in self.actual_fields
                    if '{}.'.format(excl_field) in act_field
                }
                self.actual_fields -= child_fields
            # Exclude projection itself
            self.pipeline.append({'$project': {field: 0 for field in exclude_fields}})

        return self

    def _get_stage_hierarchy_fields(self, stage, prefix=''):
        def get_fields_for_dict_value(value, prefix=''):
            if not isinstance(value, dict):
                return []
            if list(value.keys())[0].startswith('$'):
                literal = value.get('$literal')
                if literal and isinstance(literal, dict):
                    return self._get_stage_hierarchy_fields(literal, prefix)

                return [prefix]
            return self._get_stage_hierarchy_fields(value, prefix)

        paths = []
        for field, value in stage.items():
            new_prefix = f'{prefix}.{field}' if prefix else field
            if isinstance(value, list):
                for obj in value:
                    if isinstance(obj, dict):
                        paths.extend(get_fields_for_dict_value(obj, prefix=new_prefix))
                        break
                continue
            elif isinstance(value, dict):
                paths.extend(get_fields_for_dict_value(value, prefix=new_prefix))
                continue
            paths.append(new_prefix)
            continue
        return paths

    @staticmethod
    def _add_kwargs_to_args(args, kwargs):
        # Складываем kwargs в args
        if args and kwargs:
            args[-1].update(kwargs)
        elif kwargs:
            args = (kwargs,)
        elif not args:
            args = ({},)
        return args

    def group(self, *args, **kwargs):
        """
        Стадия группировки

        :param group_by: Поля группировки. Можно подавать строкой через запятую, либо списком, либо словарем
        :param first_fields: Список полей для оператора $first. Строка либо список.
        :param min_fields: Список полей для оператора $min. Строка либо список.
        :param max_fields: Список полей для оператора $max. Строка либо список.
        :param sum_fields: Список полей для оператора $sum. Строка либо список.
        :param counter_fields: Список полей для оператора $sum, количество просуммированных строк. Строка либо список.
        :param avg_fields: Список полей для оператора $avg. Строка либо список.
        :param push_fields: Список полей для оператора $push. Строка либо список.
        :param addToSet_fields: Список полей для оператора $addToSet. Строка либо список.
        :param operator_by_default: Используемый оператор для остальных полей actual_fields. По умолчанию не добавляет остальные поля в результат.
        :param exclude_fields: Использовать operator_by_default для всех, кроме перечисленных. Строка через запятую, либо список.
        :param args: Описание остальных полей группировки словарем, либо несколькими словарями.
        :param kwargs: Описание остальных полей группировки именованными аргументами.
        :return:
        """
        group_by = kwargs.pop('group_by', None)
        first_fields = kwargs.pop('first_fields', None)
        min_fields = kwargs.pop('min_fields', None)
        max_fields = kwargs.pop('max_fields', None)
        sum_fields = kwargs.pop('sum_fields', None)
        counter_fields = kwargs.pop('counter_fields', None)
        avg_fields = kwargs.pop('avg_fields', None)
        push_fields = kwargs.pop('push_fields', None)
        add_to_set_fields = kwargs.pop('addToSet_fields', None)
        operator_by_default = kwargs.pop('operator_by_default', None)
        exclude_fields = kwargs.pop('exclude_fields', False)
        if exclude_fields:
            exclude_fields = exclude_fields.split(',')

        # Инициализируем stage группировки
        stage = {}

        kwargs = _convert_names_with_underlines_to_dots(kwargs, convert_operators=True)
        # Добавляем kwargs и args в stage
        args = self._add_kwargs_to_args(args, kwargs)
        for arg in args:
            if not isinstance(arg, dict): continue
            if not arg: continue
            stage.update(arg)

        # Подготовка и обработка операторов группировки
        operators = {
            '$first': first_fields,
            '$min': min_fields,
            '$max': max_fields,
            '$sum': sum_fields,
            '$avg': avg_fields,
            '$push': push_fields,
            '$addToSet': add_to_set_fields,
        }
        for operator, fields in operators.items():
            if not fields: continue
            self._group_set_operator(stage, fields, operator)
        if counter_fields:
            self._counter_set_operator(stage, counter_fields)

        # Указываем поле группировки
        stage['_id'] = self._group_get_id(group_by)

        # Устанавливаем по дефолту $first для остальных полей
        if operator_by_default:
            parent_fields = self.get_parents(self.actual_fields, True)
            if isinstance(stage['_id'], str):
                _id_fields = pop_dollar_prefix(stage['_id'])
            else:
                _id_fields = stage['_id'].keys()
            for field in parent_fields:
                if field in chain(stage, _id_fields):
                    continue
                elif field in exclude_fields:
                    continue
                stage[field] = {operator_by_default: dollar_prefix(field)}

        # Формируем actual_fields
        if not stage['_id']:
            _id_fields = []
        elif isinstance(stage['_id'], str):
            _id_fields = ['_id']
        else:
            _id_fields = ['_id.{}'.format(key) for key in stage['_id'].keys()]

        def get_prolonged_actual_fields_for(*fields_lists):
            prolong_fields = []
            new_fields = [self._str_to_list(fields) for fields in fields_lists if fields]
            for actual_field, new_field in product(self.actual_fields, chain(*new_fields)):
                if not actual_field.startswith(new_field):
                    continue
                prolong_fields.append(actual_field)
            return prolong_fields

        prolonged_actual_fields = get_prolonged_actual_fields_for(first_fields, push_fields, add_to_set_fields)
        self._set_actual_fields(stage, _id_fields, prolonged_actual_fields)

        self.pipeline.append({'$group': stage})
        return self

    def _group_set_operator(self, stage, fields, operator):
        if not fields: return
        if isinstance(fields, str):
            fields = fields.split(',')
        fields = _convert_names_with_underlines_to_dots(fields)
        if isinstance(fields, dict):
            fields = fields.keys()
        for field in fields:
            stage[pop_dollar_prefix(field)] = {operator: dollar_prefix(field)}

    def _counter_set_operator(self, stage, fields):
        if not fields: return
        if isinstance(fields, str):
            fields = fields.split(',')
        fields = _convert_names_with_underlines_to_dots(fields)
        if isinstance(fields, dict):
            fields = fields.keys()
        for field in fields:
            stage[pop_dollar_prefix(field)] = {'$sum': 1}

    def _group_get_id(self, group_by=None):
        # Допустимая ситуация: группировка всех строк в одну
        if not group_by: return None

        # Преобразуем к списку, если строка
        if isinstance(group_by, str):
            group_by = group_by.split(',')

        # Преобразуем имена вида item__name к item.name
        group_by = _convert_names_with_underlines_to_dots(group_by)

        # Если group_by - словарь, считаем его сформированным корректно
        if isinstance(group_by, dict):
            return group_by

        # Если список, формируем словарь или одно соответсвие
        if isinstance(group_by, list):
            if len(group_by) == 1:
                return dollar_prefix(group_by[0])

            return {pop_dollar_prefix(field): dollar_prefix(field) for field in group_by}

        # Опасная ситуация, но иначе никак - пусть группирует по всем строкам
        return None

    @property
    def last_stage(self):
        if not self.pipeline:
            return None
        return LastStage(self.pipeline)

    def revert_last_stage(self):
        if not self.pipeline:
            return
        self.pipeline = self.pipeline[:-1]

    def get_first(self, default=None, **kwargs):
        # Limit pipeline to 1 document
        self.limit(1)
        # Get first document to include statistics_today lately
        first_doc = next(self.aggregate(**kwargs), default)
        # Undo limit
        self.revert_last_stage()
        return first_doc

    def get_count(self, **kwargs):
        return self.count(**kwargs)


class LastStage():

    def __init__(self, pipeline):
        if not pipeline:
            raise Exception("Последнего этапа в этой агрегации не существует.")
        self.__data = pipeline[-1]

    @property
    def name(self):
        return list(self.__data.keys())[0]

    @name.setter
    def name(self, value):
        self.__data[value] = self.__data.pop(self.name)

    @property
    def statement(self):
        return self.__data[self.name]

    @statement.setter
    def statement(self, value):
        self.__data[self.name] = value

    def __repr__(self):
        return str(self.__data)


if __name__ == "__main__":
    import logging
    import os

    from dotenv import load_dotenv

    load_dotenv()

    mongo_connection = os.environ.get('MONGODB_HOST', 'mongodb://localhost:27017/test')
    import pymongo
    from datetime import datetime
    from dateutil.relativedelta import relativedelta

    logger = logging.getLogger()
    client = pymongo.MongoClient(mongo_connection)
    db = client.get_default_database()
    today = datetime.now()
    yesterday = today - relativedelta(days=1)

    pipeline = MongoAggregation(collection=db.action)
    pipeline.match(
        {'date': {'$gte': yesterday, '$lt': today}},
        completed=True,
        _cls={'$in': [
            'Action.Order', 'Action.StorageIncome', 'Action.StorageCancellation', 'Action.StorageMovementOutcome'
        ]},
    ).smart_project(
        'transactions.amount,_cls', '_id'
    ).append([
        {'$project': {
            '_cls': 1,
            'date': 1,
            'transactions.amount': 1,
            'transactions.cashbox': 1,
        }},
    ]).project(
        {'transactions.amount': 1, 'transactions.cashbox': 1}, _cls=1, date=1
    ).project(
        transactions=1, _cls=1, date=1
    )
    cursor = pipeline.aggregate()

    for doc in cursor:
        print(doc)
        break


# coding=utf-8


class MongoMatchFilter(dict):
    """
    Класс, представляющий собой словарь raw-запроса к mongo.
    Реализует запрет повторной установки ключа $or:
    Поскольку для запроса используется словарь типа dict, возможно неожидаемое переопределение ключа $or.
    """
    def __setitem__(self, key, value):
        """
        Устанваливает ключи в словарь запроса mongo.
        Если устанавливается больше одного ключа $or, вызывается исключение.
        Для корректного использования нескольких $or-условий одновременно, используйте метод and_or().
        :param key: ключ словаря
        :param value: значение словаря по ключу
        :return:
        """
        if key == '$or' and key in self:
            raise KeyError('Невозможно установить ключ $or однозначно, т.к. ключ уже существует. '
                           'При повторном назначении ключа возможно его переопределение. '
                           'Для корректной установки второго $or воспользуйтесь методом and_or().')

        super(MongoMatchFilter, self).__setitem__(key, value)

    def and_(self, *args, **kwargs):
        """
        Добавляет фильтр используя опертор and. Позволяет избежать ситуации, с дублированием ключей.
        При необходимости создает ключ $and.
        :param value: Устанавливаемое значение
        :return: список условия $or
        """
        # Складываем все в args
        args = list(args)
        if kwargs:
            args.append(kwargs)

        for arg in args:
            for key in arg:
                if key in self:
                    break
            else:
                self.update(arg)
                continue
            self.setdefault('$and', [])
            self['$and'].append(arg)

    def and_nor(self, *args):
        """
        Добавляет в mongo-запрос ключ $nor.
        Во избежание перезаписи ключа $nor, ключи $nor записываются в список условий $and.
        :param args: Список $nor
        :return: список условия $or
        """
        self.and_or(*args, negative=True)

    def and_or(self, *args, negative=False):
        """
        Добавляет в mongo-запрос ключ $or.
        Во избежание перезаписи ключа $or, ключи $or записываются в список условий $and.
        :param args: Список $or
        :param negative: Флаг для использования $nor
        :return: список условия $or
        """
        key = '$nor' if negative else '$or'
        or_dict = MongoMatchFilter()
        or_dict[key] = args[0] if len(args) == 1 and isinstance(args[0], list) else list(args)
        if len(or_dict[key]) == 1 and not negative:
            or_dict = or_dict[key][0]
        self.and_(or_dict)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from copy import deepcopy

import six

from mongo_aggregation.patterns import _convert_names_with_underlines_to_dots, dollar_prefix, _list_dollar_prefix


def field_is_specified(field):
    return {'$cond': [field, True, False]}


def project_set_value(expression):
    #TODO переделать на $literal если получится
    return {'$cond': [True, expression, '']}


def filter_(input, condition, as_field=''):
    """Selects a subset of an array to return based on the specified condition.
    Returns an array with only those elements that match the condition.
    The returned elements are in the original order.
    From mongo version 3.2."""
    filter_stage = {'input': input, 'cond': condition}
    if as_field:
        filter_stage['as'] = as_field
    return {'$filter': filter_stage}


def round_half_up(expression, signs_after_dot=0):
    abs_expression = {'$abs': expression}
    sign = {'$cond': [{'$gt': [expression, 0]}, 1, -1]}
    if signs_after_dot:
        rounded_abs = {'$divide': [
            {'$trunc': {'$add': [{'$multiply': [abs_expression, pow(10, signs_after_dot)]}, 0.5]}},
            pow(10, signs_after_dot)
        ]}
    else:
        rounded_abs = {'$trunc': {'$add': [abs_expression, 0.5]}}
    return {'$multiply': [rounded_abs, sign]}


def day_start(field):
    field = dollar_prefix(field)
    return {'$subtract': [
        field,
        {'$add': [
            {'$multiply': [{'$hour': field}, 3600000]},
            {'$multiply': [{'$minute': field}, 60000]},
            {'$multiply': [{'$second': field}, 1000]},
            {'$millisecond': field}
        ]}
    ]}


def month_start(field):
    field = dollar_prefix(field)
    return {'$subtract': [
        field,
        {'$add': [
            {'$multiply': [{'$subtract': [{'$dayOfMonth': field}, 1]}, 86400000]},
            {'$multiply': [{'$hour': field}, 3600000]},
            {'$multiply': [{'$minute': field}, 60000]},
            {'$multiply': [{'$second': field}, 1000]},
            {'$millisecond': field}
        ]}
    ]}


def lookup_unwind(collection, local_field='_id', as_field='', foreign_field='_id', preserveNullAndEmptyArrays=True):
    if not as_field:
        as_field = local_field
    pipeline = [
        lookup(collection, local_field, as_field, foreign_field),
        unwind(as_field, preserveNullAndEmptyArrays)
    ]
    return pipeline


def lookup(collection, local_field='_id', as_field='', foreign_field='_id'):
    if not as_field:
        as_field = local_field
    return {"$lookup": {
        'from': collection,
        'foreignField': foreign_field,
        'localField': local_field,
        'as': as_field,
    }}


def unwind(field, preserveNullAndEmptyArrays=False):
    field = dollar_prefix(field)
    if preserveNullAndEmptyArrays:
        return {'$unwind': {'path': field, 'preserveNullAndEmptyArrays': True}}
    else:
        return {'$unwind': field}


def date_to_string(field, date_format="%Y-%m-%d %H:%M:%S:%L"):
    field = dollar_prefix(field)
    return {'$dateToString': {'format': date_format, 'date': field}}


def set_value(expression):
    return {'$literal': expression}


def merge_objects(*args, **kwargs):
    """Combines multiple documents into a single document.
    $mergeObjects operator implementation.
    From MongoDB version 3.6."""
    # Put everything into args
    args = list(args)
    args = [dollar_prefix(arg) if isinstance(arg, six.string_types) else arg for arg in args]
    if kwargs:
        kwargs = _convert_names_with_underlines_to_dots(kwargs, convert_operators=True)
        args.append(kwargs)
    return {'$mergeObjects': args}


def hide_fields(fields_to_hide, visible_fields, concealer_field='_id'):
    if isinstance(fields_to_hide, str):
        fields_to_hide = fields_to_hide.split(',')
    if isinstance(visible_fields, str):
        visible_fields = visible_fields.split(',')

    projection = {
        concealer_field: {field: dollar_prefix(field) for field in fields_to_hide},
    }
    projection[concealer_field][concealer_field] = dollar_prefix(concealer_field)
    projection.update({field: 1 for field in visible_fields})
    return {'$project': projection}


def unhide_fields(hided_fields, visible_fields, concealer_field='_id'):
    if isinstance(hided_fields, str):
        hided_fields = hided_fields.split(',')
    if isinstance(visible_fields, str):
        visible_fields = visible_fields.split(',')

    projection = {
        field: '{}.{}'.format(dollar_prefix(concealer_field), field)
        for field in hided_fields
    }
    projection.update({field: 1 for field in visible_fields})
    if concealer_field == '_id' and concealer_field not in hided_fields:
        projection[concealer_field] = 0
    return {'$project': projection}


def switch(cases_list, last_as_final=False, final_else=''):
    if not cases_list: return final_else
    cases = deepcopy(cases_list)
    case, value = cases.pop(0)
    if not cases and last_as_final: return {'$literal': value}
    else_expr = switch(cases, last_as_final, final_else)
    return {'$cond': [case, value, else_expr]}


def switch_compare(field, cases_list, compare_method='$eq', final_else=''):
    if not cases_list: return final_else
    cases = deepcopy(cases_list)
    case, value = cases.pop(0)
    else_expr = switch_compare(field, cases, compare_method, final_else)
    return {'$cond': [{compare_method: [field, case]}, value, else_expr]}


def if_null(field, value):
    """$ifNull implementation."""
    return {'$ifNull': [dollar_prefix(field), value]}


def cond(condition, then_value, else_value=0):
    """$cond implementation."""
    if isinstance(condition, str):
        condition = dollar_prefix(condition)
    return {"$cond": [condition, then_value, else_value]}


def concat(*args):
    """$concat implementation."""
    return {'$concat': args}


def multiply(*args):
    """$multiply implementation."""
    return {'$multiply': _list_dollar_prefix(*args)}


def divide(first_expression, second_expression):
    """$divide implementation."""
    return {'$divide': _list_dollar_prefix(first_expression, second_expression)}


def sum_(*args):
    args = _list_dollar_prefix(*args)
    if len(args) == 1:
        return {'$sum': args[0]}
    return {'$sum': _list_dollar_prefix(*args)}


def in_(expression, array_expression):
    """$multiply implementation.

    >>> in_('a', 'b')
    {'$in': ['$a', '$b']}
    >>> in_('$a', 'b')
    {'$in': ['$a', '$b']}
    >>> in_('a', '$b')
    {'$in': ['$a', '$b']}
    >>> in_('a', ['b', 'c'])
    {'$in': ['$a', ['b', 'c']]}
    """
    return {'$in': _list_dollar_prefix(expression, array_expression)}


if __name__ == '__main__':

    def _test_module():
        import doctest
        result = doctest.testmod()
        if not result.failed:
            print(f"{result.attempted} passed and {result.failed} failed.\nTest passed.")

    _test_module()



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="mongo-aggregation",
    version='1.0.12',
    description="Python MongoDB aggregation ORM",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/egorgvo/mongo_aggregation",
    author="Egor Gvo",
    author_email="work.egvo@ya.ru",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    packages=["mongo_aggregation"],
)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from copy import copy

import six


def dollar_prefix(field):
    return field if field.startswith('$') else '${}'.format(field)


def _list_dollar_prefix(*args):
    args = list(args)
    for i, arg in enumerate(args):
        if not isinstance(arg, str):
            continue
        args[i] = dollar_prefix(arg)
    return args


def pop_dollar_prefix(field):
    return field[1:] if field.startswith('$') else field


def regex(pattern, options='', i=False, m=False, x=False, s=False, extra_fields=None):
    """Provides regular expression capabilities for pattern matching strings in queries."""
    expression = {'$regex': pattern}
    if not options:
        options = f'{"i" if i else ""}{"m" if m else ""}{"x" if x else ""}{"s" if s else ""}'
    if options:
        expression['$options'] = options
    if extra_fields:
        expression.update(extra_fields)
    return expression


def _and(*args):
    """Deprecated. Use and_ instead."""
    return and_(*args)


def and_(*args, **kwargs):
    """Performs a logical AND operation on an array of one or more expressions
    (e.g. <expression1>, <expression2>, etc.) and selects the documents that
    satisfy all the expressions in the array."""
    # Складываем все в args
    args = list(args)
    if kwargs:
        kwargs = _convert_names_with_underlines_to_dots(kwargs, convert_operators=True)
        kwargs = [{key: value} for key, value in kwargs.items()]
        args.extend(kwargs)
    return {'$and': args}


def or_(*args, **kwargs):
    """Performs a logical OR operation on an array of two or more expressions
    and selects the documents that satisfy at least one of the expressions."""
    # Складываем все в args
    args = list(args)
    if kwargs:
        kwargs = _convert_names_with_underlines_to_dots(kwargs, convert_operators=True)
        kwargs = [{key: value} for key, value in kwargs.items()]
        args.extend(kwargs)
    return {'$or': args}


def eq(first_expression, second_expression=True):
    return {'$eq': [first_expression, second_expression]}


def gt(first_expression, second_expression=None):
    if second_expression is None:
        return {'$gt': first_expression}
    return {'$gt': [first_expression, second_expression]}


def gte(first_expression, second_expression=None):
    if second_expression is None:
        return {'$gte': first_expression}
    return {'$gte': [first_expression, second_expression]}


def lte(first_expression, second_expression=None):
    if second_expression is None:
        return {'$lte': first_expression}
    return {'$lte': [first_expression, second_expression]}


def ne(first_expression, second_expression=True):
    if second_expression is True:
        return {'$not': [first_expression]}
    return {'$ne': [first_expression, second_expression]}


def field_in(field, values_list):
    field = dollar_prefix(field)
    return {'$or': [
        eq(field, value)
        for value in values_list
    ]}


def optimized_in(value, negative=False, operator_is_required=False):
    """
    Реализует оптимальный фильтр оператора $in. Если значение в списке единственное, то используется оператор $eq.
    Также возможно использование с отрицанием: $nin и $ne.
    Для оператора $eq по умолчанию возвращается только значение, потому что этого достаточно для обычного фильтра.
    Для возврата значения с оператором используйте operator_is_required в значении True.
    :param value: Список значений фильтра
    :param negative: Флаг, использовать ли $nin и $ne вместо $in и $eq
    :param operator_is_required: Флаг, обязательно ли возвращать оператор со значением
    :return: Значение или словарь {оператор: значение}
    >>> optimized_in(['1'])
    '1'
    >>> optimized_in(['1'], operator_is_required=True)
    {'$eq': '1'}
    >>> optimized_in(['1'], negative=True)
    {'$ne': '1'}
    >>> optimized_in(['1', '2'])
    {'$in': ['1', '2']}
    >>> optimized_in(['1', '2'], negative=True)
    {'$nin': ['1', '2']}
    """
    if len(value) == 1:
        value = value[0]
        operator = '$ne' if negative else '$eq' if operator_is_required else ''
    else:
        operator = '$nin' if negative else '$in'

    if operator:
        return {operator: value}
    return value


def field_exists(field):
    """
    >>> field_exists('field')
    {'$or': [{'$eq': ['$field', None]}, {'$gt': ['$field', None]}]}
    >>> field_exists('$field')
    {'$or': [{'$eq': ['$field', None]}, {'$gt': ['$field', None]}]}
    """
    field = dollar_prefix(field)
    return {'$or': [
        {'$eq': [field, None]},
        {'$gt': [field, None]},
    ]}


def _convert_names_with_underlines_to_dots(args, convert_operators=False):
    """Проверяем обращение к полям объекта с помощью __ в словаре"""
    for i, arg in enumerate(copy(args)):
        if '__' not in arg:
            continue
        elif arg.startswith('__') or arg.endswith('__'):
            continue
        replacement = arg.replace('__', '.')
        if isinstance(args, list):
            args.pop(i)
            args.insert(i, replacement)
        else:
            if convert_operators:
                potential_operator = replacement[replacement.rfind('.')+1:]
                field = replacement[:replacement.rfind('.')]
                if potential_operator in ['eq', 'ne', 'gt', 'gte', 'lt', 'lte', 'in', 'nin', 'push']:
                    args.setdefault(field, {})
                    args[field][dollar_prefix(potential_operator)] = args.pop(arg)
                    continue
                elif potential_operator in ['regex', 'iregex', 'icontains', 'contains']:
                    ignore_case = potential_operator.startswith('i')
                    args.setdefault(field, {})
                    args[field].update(regex(args.pop(arg), i=ignore_case))
                    continue
            args[replacement] = args.pop(arg)
    return args


def obj(*args, **kwargs):
    """
    Returns dictionary with specified fields as keys and values. See usage examples in tests below.

    >>> obj('name,description')
    {'name': '$name', 'description': '$description'}
    >>> obj('name', 'description')
    {'name': '$name', 'description': '$description'}
    >>> obj('name,description', id='$code')
    {'name': '$name', 'description': '$description', 'id': '$code'}
    """
    fields = [f for arg in args for f in arg.split(',') if isinstance(arg, six.string_types)]
    for arg in args:
        if isinstance(arg, dict):
            kwargs.update(arg)
    kwargs = _convert_names_with_underlines_to_dots(kwargs, convert_operators=True)
    result = {pop_dollar_prefix(field): dollar_prefix(field) for field in fields}
    result.update(kwargs)
    return result


if __name__ == '__main__':

    def _test_module():
        import doctest
        result = doctest.testmod()
        if not result.failed:
            print(f"{result.attempted} passed and {result.failed} failed.\nTest passed.")

    _test_module()
